#include "BasicBlock.h"
#include <algorithm>
#include <vector>
#include "Function.h"

using namespace std;
extern FILE* yyout;

// insert the instruction to the front of the basicblock.
void BasicBlock::insertFront(Instruction* inst) {
    insertBefore(inst, head->getNext());
}

// insert the instruction to the back of the basicblock.
void BasicBlock::insertBack(Instruction* inst) {
    insertBefore(inst, head);
}

// insert the instruction dst before src.
void BasicBlock::insertBefore(Instruction* dst, Instruction* src) {
    // Todo
    src->getPrev()->setNext(dst);
    dst->setPrev(src->getPrev());

    dst->setNext(src);
    src->setPrev(dst);

    dst->setParent(this);
}

// insert the instruction dst after src.
void BasicBlock::insertAfter(Instruction* dst, Instruction* src) {
    dst->setNext(src->getNext());
    src->getNext()->setPrev(dst);

    dst->setPrev(src);
    src->setNext(dst);

    dst->setParent(this);
}

// remove the instruction from intruction list.
void BasicBlock::remove(Instruction* inst) {
    inst->getPrev()->setNext(inst->getNext());
    inst->getNext()->setPrev(inst->getPrev());
}

void BasicBlock::output() const {
    fprintf(yyout, "B%d:", no);

    if (!pred.empty()) {
        fprintf(yyout, "%*c; preds = %%B%d", 32, '\t', pred[0]->getNo());
        for (auto i = pred.begin() + 1; i != pred.end(); i++)
            fprintf(yyout, ", %%B%d", (*i)->getNo());
    }
    fprintf(yyout, "\n");
    for (auto i = head->getNext(); i != head; i = i->getNext())
        i->output();
}

void BasicBlock::addSucc(BasicBlock* bb, bool first) {
    if (first)
        succ.insert(succ.begin(), bb);
    else
        succ.push_back(bb);
}

// remove the successor basicclock bb.
void BasicBlock::removeSucc(BasicBlock* bb) {
    auto iter = std::find(succ.begin(), succ.end(), bb);
    if (iter != succ.end())
        succ.erase(iter);
}

void BasicBlock::removeSuccFromEnd(BasicBlock* bb) {
    succ.erase((std::find(succ.rbegin(), succ.rend(), bb) + 1).base());
}

void BasicBlock::addPred(BasicBlock* bb) {
    pred.push_back(bb);
}

// remove the predecessor basicblock bb.
void BasicBlock::removePred(BasicBlock* bb) {
    auto iter = std::find(pred.begin(), pred.end(), bb);
    if (iter != pred.end())
        pred.erase(iter);
}

void BasicBlock::removePredFromEnd(BasicBlock* bb) {
    auto iter = std::find(pred.rbegin(), pred.rend(), bb);
    if (iter != pred.rend())
        pred.erase((iter + 1).base());
}

void BasicBlock::genMachineCode(AsmBuilder* builder) {
    auto cur_func = builder->getFunction();
    auto cur_block = new MachineBlock(cur_func, no);
    builder->setBlock(cur_block);
    for (auto i = head->getNext(); i != head; i = i->getNext()) {
        i->genMachineCode(builder);
    }
    cur_func->InsertBlock(cur_block);
}

BasicBlock::BasicBlock(Function* f) {
    this->no = SymbolTable::getLabel();
    f->insertBlock(this);
    parent = f;
    head = new DummyInstruction();
    head->setParent(this);
    mark = false;
    lastAlloc = nullptr;
}

BasicBlock::~BasicBlock() {
    Instruction* inst;
    inst = head->getNext();
    while (inst != head) {
        Instruction* t;
        t = inst;
        inst = inst->getNext();
        delete t;
    }
    for (auto& bb : pred)
        bb->removeSucc(this);
    for (auto& bb : succ)
        bb->removePred(this);
    parent->remove(this);
}

void BasicBlock::cleanSucc() {
    for (auto i : succ)
        i->removePred(this);
    vector<BasicBlock*>().swap(succ);
}

void BasicBlock::cleanMark() {
    auto inst = head->getNext();
    while (inst != head) {
        inst->unsetMark();
        inst = inst->getNext();
    }
}
void BasicBlock::insertPhiInstruction(Operand* dst) {
    Instruction* i = new PhiInstruction(dst);
    insertFront(i);
}

void BasicBlock::deleteBack(int num) {
    while (num--) {
        remove(head->getPrev());
    }
}

bool BasicBlock::isBefore(Instruction* a, Instruction* b) {
    if (a->getParent() != this)
        return true;
    assert(b->getParent() == this);
    auto temp = a;
    while (temp != head) {
        if (temp == b)
            return true;
        temp = temp->getNext();
    }
    return false;
}

void BasicBlock::replaceIns(Instruction* old, Instruction* new_) {
    old->getPrev()->setNext(new_);
    new_->setPrev(old->getPrev());
    old->getNext()->setPrev(new_);
    new_->setNext(old->getNext());
}

void BasicBlock::addAlloc(Instruction* alloc) {
    assert(alloc->isAlloc());
    if (!lastAlloc) {
        insertFront(alloc);
        lastAlloc = alloc;
    } else {
        insertAfter(alloc, lastAlloc);
        lastAlloc = alloc;
    }
}